==> ngautonml-0.4.1b2.tar.gz: ngautonml-0.4.1b2/ngautonml/communicators/sockets_communicator_test.py <==
'''Tests for sockets_communicator.py.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring


from queue import Queue
import socket
from typing import Dict, Set


from ..config_components.distributed_config import DistributedConfig
from ..neighbor_manager.event import NewNeighbor, Recv
from ..neighbor_manager.node_id import NodeID
from ..wrangler.base_port import BasePort

from .impl.ip_endpoint import IPEndpoint
from .sockets_communicator import SocketsCommunicator

base_port = BasePort(12345)

LOOPBACK = '127.0.0.1'

RECV_UNIT_PORT_0 = base_port.next()
RECV_UNIT_PORT_1 = base_port.next()


def test_receive() -> None:
    '''SocketsCommunicator listens on node 0, node 1 talks on a socket'''
    known_neighbors: Set[NodeID] = set()
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    (0, (LOOPBACK, RECV_UNIT_PORT_0)),
                    (1, (LOOPBACK, RECV_UNIT_PORT_1)),
                ]
            }
        }
    })
    dut = SocketsCommunicator(
        my_id=NodeID(0),
        distributed=distributed,
        known_neighbors=known_neighbors
    )
    queue: Queue = Queue()
    dut.start(queue=queue)
    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, RECV_UNIT_PORT_1))
    sock.sendto(b'Hello, world!', (LOOPBACK, RECV_UNIT_PORT_0))
    sock.close()
    try:
        first_event = queue.get(block=True, timeout=0.5)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(1)
        assert known_neighbors == {NodeID(1)}

        second_event = queue.get(block=True, timeout=0.5)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(1)
        assert second_event.payload == b'Hello, world!'
    finally:
        dut.stop()


SEND_UNIT_PORT_0 = base_port.next()
SEND_UNIT_PORT_1 = base_port.next()

SEND_UNIT_NODES_AND_ENDPOINTS = [
    (NodeID(0), IPEndpoint(LOOPBACK, SEND_UNIT_PORT_0)),
    (NodeID(1), IPEndpoint(LOOPBACK, SEND_UNIT_PORT_1))
]


def test_send() -> None:
    '''SocketsCommunicator talks on node 1, node 0 listens on a socket'''
    known_neighbors: Set[NodeID] = set()
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    (0, (LOOPBACK, SEND_UNIT_PORT_0)),
                    (1, (LOOPBACK, SEND_UNIT_PORT_1)),
                ]
            }
        }
    })

    dut = SocketsCommunicator(
        my_id=NodeID(1),
        known_neighbors=known_neighbors,
        distributed=distributed
    )
    queue: Queue = Queue()
    dut.start(queue=queue)

    listener = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    listener.bind((LOOPBACK, SEND_UNIT_PORT_0))

    try:

        dut.send(dest_id=NodeID(0), payload=b'Hello World!')

        payload, addr = listener.recvfrom(1024)
        assert payload == b'Hello World!'
        assert addr == (LOOPBACK, SEND_UNIT_PORT_1)
    finally:
        listener.close()
        dut.stop()


INTEGRATION_PORT_0 = base_port.next()
INTEGRATION_PORT_1 = base_port.next()

INTEGRATION_NODES_AND_ENDPOINTS = [
    (NodeID(0), IPEndpoint(LOOPBACK, INTEGRATION_PORT_0)),
    (NodeID(1), IPEndpoint(LOOPBACK, INTEGRATION_PORT_1))
]


def test_exchange() -> None:
    '''2 SocketsCommunicator on different nodes talk and listen to each other.'''

    known_neighbors0: Set[NodeID] = set()
    known_neighbors1: Set[NodeID] = set()

    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    (0, (LOOPBACK, INTEGRATION_PORT_0)),
                    (1, (LOOPBACK, INTEGRATION_PORT_1)),
                ]
            }
        }
    })

    dut0 = SocketsCommunicator(
        my_id=NodeID(0),
        known_neighbors=known_neighbors0,
        distributed=distributed
    )
    dut1 = SocketsCommunicator(
        my_id=NodeID(1),
        known_neighbors=known_neighbors1,
        distributed=distributed
    )

    queue0: Queue = Queue()
    queue1: Queue = Queue()

    dut0.start(queue=queue0)
    dut1.start(queue=queue1)

    try:
        dut0.send(dest_id=NodeID(1), payload=b'Hello, Node 1!')
        first_event = queue1.get(block=True, timeout=0.5)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(0)
        assert known_neighbors1 == {NodeID(0)}

        second_event = queue1.get(block=True, timeout=0.5)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(0)
        assert second_event.payload == b'Hello, Node 1!'

        dut1.send(dest_id=NodeID(0), payload=b'Nice to meet you, Node 0 :)')

        first_event = queue0.get(block=True, timeout=0.5)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(1)
        assert known_neighbors0 == {NodeID(1)}

        second_event = queue0.get(block=True, timeout=0.5)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(1)
        assert second_event.payload == b'Nice to meet you, Node 0 :)'

    finally:
        dut0.stop()
        dut1.stop()


SENDALL_PORT_0 = base_port.next()
SENDALL_PORT_1 = base_port.next()
SENDALL_PORT_2 = base_port.next()
SENDALL_PORT_3 = base_port.next()
SENDALL_PORT_4 = base_port.next()


SENDALL_NODES_AND_ENDPOINTS = [
    (NodeID(0), IPEndpoint(LOOPBACK, SENDALL_PORT_0)),
    (NodeID(1), IPEndpoint(LOOPBACK, SENDALL_PORT_1)),
    (NodeID(2), IPEndpoint(LOOPBACK, SENDALL_PORT_2)),
    (NodeID(3), IPEndpoint(LOOPBACK, SENDALL_PORT_3)),
    (NodeID(4), IPEndpoint(LOOPBACK, SENDALL_PORT_4)),
]


def test_sendall() -> None:
    '''One node sends to a bunch of neighbors.'''

    known_neighbors: Dict[NodeID, Set[NodeID]] = {}
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    (0, (LOOPBACK, SENDALL_PORT_0)),
                    (1, (LOOPBACK, SENDALL_PORT_1)),
                    (2, (LOOPBACK, SENDALL_PORT_2)),
                    (3, (LOOPBACK, SENDALL_PORT_3)),
                    (4, (LOOPBACK, SENDALL_PORT_4)),
                ]
            }
        }
    })

    dut: Dict[NodeID, SocketsCommunicator] = {}
    queue: Dict[NodeID, Queue] = {}
    for node_id in range(0, 5):
        node = NodeID(node_id)
        known_neighbors[node] = set()
        queue[node] = Queue()
        dut[node] = SocketsCommunicator(
            my_id=node,
            known_neighbors=known_neighbors[node],
            distributed=distributed)
        dut[node].start(queue=queue[node], timeout=0.5)

    for node_id in range(1, 5):
        node = NodeID(node_id)
        # We need to send an initial message to establish our neighborliness.
        dut[node].send(NodeID(0), f'Hello, I am {node}.'.encode())

    try:
        # Read a NewNeighbor and a Recv for each neighbor.
        # The pairs of requests are randomly interleaved due to threading,
        # so we don't bother checking what they are.
        for _ in range(1, 5):
            _ = queue[NodeID(0)].get()
            _ = queue[NodeID(0)].get()

        # This is what we are testing.
        dut[NodeID(0)].send_all(payload=b'Hello, All Nodes!')

        # Confirm that the test worked.
        for node_id in range(1, 5):
            node = NodeID(node_id)

            first_event = queue[node].get(block=True, timeout=0.5)
            assert isinstance(first_event, NewNeighbor)
            assert first_event.neighbor == NodeID(0)
            assert known_neighbors[node] == {NodeID(0)}

            second_event = queue[node].get(block=True, timeout=0.5)
            assert isinstance(second_event, Recv)
            assert second_event.neighbor == NodeID(0)
            assert known_neighbors[node] == {NodeID(0)}
            assert second_event.payload == b'Hello, All Nodes!'

    finally:
        for node_id in range(0, 5):
            node = NodeID(node_id)
            dut[node].stop()


DYNAMIC_DISCOVERY_PORT_0 = base_port.next()
DYNAMIC_DISCOVERY_PORT_1 = base_port.next()
DYNAMIC_DISCOVERY_PORT_2 = base_port.next()
DYNAMIC_DISCOVERY_PORT_3 = base_port.next()
DYNAMIC_DISCOVERY_PORT_4 = base_port.next()


DYNAMIC_DISCOVERY_NODES_AND_ENDPOINTS = [
    (NodeID(0), IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_0)),
]

DYNAMIC_DISCOVERY_ENDPOINTS = [
    IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_0),
    IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_1),
    IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_2),
    IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_3),
    IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_4),
]


def test_dynamic_discovery() -> None:
    '''One node sends to a bunch of neighbors.'''

    known_neighbors: Dict[NodeID, Set[NodeID]] = {}
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    (0, (LOOPBACK, DYNAMIC_DISCOVERY_PORT_0)),
                    (1, (LOOPBACK, DYNAMIC_DISCOVERY_PORT_1)),
                    (2, (LOOPBACK, DYNAMIC_DISCOVERY_PORT_2)),
                    (3, (LOOPBACK, DYNAMIC_DISCOVERY_PORT_3)),
                    (4, (LOOPBACK, DYNAMIC_DISCOVERY_PORT_4)),
                ]
            }
        }
    })

    dut: Dict[NodeID, SocketsCommunicator] = {}
    queue: Dict[NodeID, Queue] = {}
    for node_id in range(0, 5):
        node = NodeID(node_id)
        known_neighbors[node] = set()
        queue[node] = Queue()
        dut[node] = SocketsCommunicator(
            my_id=node,
            known_neighbors=known_neighbors[node],
            distributed=distributed)
        dut[node].start(queue=queue[node], timeout=0.5)

    # This is the test. Everybody introduces themselves.
    for node_id in range(1, 5):
        node = NodeID(node_id)
        dut[node].send(NodeID(0), f'Hello, I am {node}.'.encode())

    # Confirm that the test worked.
    try:
        # Read a NewNeighbor and a Recv for each neighbor.
        # The pairs of requests are randomly interleaved due to threading,
        # so we don't bother checking what they are.
        # This is needed to make sure every message has been recieved before
        #   checking known_neighbors.
        for _ in range(1, 5):
            _ = queue[NodeID(0)].get()
            _ = queue[NodeID(0)].get()

        assert known_neighbors[NodeID(0)] == {
            NodeID(1),
            NodeID(2),
            NodeID(3),
            NodeID(4),
        }

        known_endpoints = {
            dut[NodeID(0)].lookup_endpoint(node_id).port
            for node_id in known_neighbors[NodeID(0)]
        }
        assert known_endpoints == {
            DYNAMIC_DISCOVERY_PORT_1,
            DYNAMIC_DISCOVERY_PORT_2,
            DYNAMIC_DISCOVERY_PORT_3,
            DYNAMIC_DISCOVERY_PORT_4,
        }

    finally:
        for node_id in range(0, 5):
            node = NodeID(node_id)
            dut[node].stop()

==> ngautonml-0.4.3b4.tar.gz: ngautonml-0.4.3b4/ngautonml/communicators/sockets_communicator_test.py <==
'''Tests for sockets_communicator.py.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring
# pylint: disable=duplicate-code, redefined-outer-name


from queue import Queue
import socket
from typing import Dict, Set


from ..config_components.distributed_config import DistributedConfig
from ..neighbor_manager.event import NewNeighbor, Recv
from ..neighbor_manager.node_id import NodeID
from ..wrangler.base_port import BasePort

from .impl.communicator_auto import CommunicatorCatalogAuto
from .impl.ip_endpoint import IPEndpoint
from .sockets_communicator import SocketsCommunicator

_ = CommunicatorCatalogAuto()

base_port = BasePort()

LOOPBACK = '127.0.0.1'

RECV_UNIT_PORT_0 = base_port.next()
RECV_UNIT_PORT_1 = base_port.next()


def test_receive() -> None:
    '''SocketsCommunicator listens on node 0, node 1 talks on a socket'''
    known_neighbors: Set[NodeID] = set()
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    (0, (LOOPBACK, RECV_UNIT_PORT_0)),
                    (1, (LOOPBACK, RECV_UNIT_PORT_1)),
                ]
            }
        }
    })
    dut = SocketsCommunicator(
        my_id=NodeID(0),
        distributed=distributed,
        known_neighbors=known_neighbors
    )
    queue: Queue = Queue()
    dut.start(queue=queue)
    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, RECV_UNIT_PORT_1))
    sock.sendto(b'Hello, world!', (LOOPBACK, RECV_UNIT_PORT_0))
    sock.close()
    try:
        first_event = queue.get(block=True, timeout=0.5)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(1)
        assert known_neighbors == {NodeID(1)}

        second_event = queue.get(block=True, timeout=0.5)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(1)
        assert second_event.payload == b'Hello, world!'
    finally:
        dut.stop()


SEND_UNIT_PORT_0 = base_port.next()
SEND_UNIT_PORT_1 = base_port.next()

SEND_UNIT_NODES_AND_ENDPOINTS = [
    (NodeID(0), IPEndpoint(LOOPBACK, SEND_UNIT_PORT_0)),
    (NodeID(1), IPEndpoint(LOOPBACK, SEND_UNIT_PORT_1))
]


def test_send() -> None:
    '''SocketsCommunicator talks on node 1, node 0 listens on a socket'''
    known_neighbors: Set[NodeID] = set()
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    (0, (LOOPBACK, SEND_UNIT_PORT_0)),
                    (1, (LOOPBACK, SEND_UNIT_PORT_1)),
                ]
            }
        }
    })

    dut = SocketsCommunicator(
        my_id=NodeID(1),
        known_neighbors=known_neighbors,
        distributed=distributed
    )
    queue: Queue = Queue()
    dut.start(queue=queue)

    listener = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    listener.bind((LOOPBACK, SEND_UNIT_PORT_0))

    try:

        dut.send(dest_ids=[NodeID(0)], payload=b'Hello World!')

        payload, addr = listener.recvfrom(1024)
        assert payload == b'Hello World!'
        assert addr == (LOOPBACK, SEND_UNIT_PORT_1)
    finally:
        listener.close()
        dut.stop()


INTEGRATION_PORT_0 = base_port.next()
INTEGRATION_PORT_1 = base_port.next()

INTEGRATION_NODES_AND_ENDPOINTS = [
    (NodeID(0), IPEndpoint(LOOPBACK, INTEGRATION_PORT_0)),
    (NodeID(1), IPEndpoint(LOOPBACK, INTEGRATION_PORT_1))
]


def test_exchange() -> None:
    '''2 SocketsCommunicator on different nodes talk and listen to each other.'''

    known_neighbors0: Set[NodeID] = set()
    known_neighbors1: Set[NodeID] = set()

    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    (0, (LOOPBACK, INTEGRATION_PORT_0)),
                    (1, (LOOPBACK, INTEGRATION_PORT_1)),
                ]
            }
        }
    })

    dut0 = SocketsCommunicator(
        my_id=NodeID(0),
        known_neighbors=known_neighbors0,
        distributed=distributed
    )
    dut1 = SocketsCommunicator(
        my_id=NodeID(1),
        known_neighbors=known_neighbors1,
        distributed=distributed
    )

    queue0: Queue = Queue()
    queue1: Queue = Queue()

    dut0.start(queue=queue0)
    dut1.start(queue=queue1)

    try:
        dut0.send(dest_ids=[NodeID(1)], payload=b'Hello, Node 1!')
        first_event = queue1.get(block=True, timeout=0.5)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(0)
        assert known_neighbors1 == {NodeID(0)}

        second_event = queue1.get(block=True, timeout=0.5)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(0)
        assert second_event.payload == b'Hello, Node 1!'

        dut1.send(dest_ids=[NodeID(0)], payload=b'Nice to meet you, Node 0 :)')

        first_event = queue0.get(block=True, timeout=0.5)
        assert isinstance(first_event, NewNeighbor)
        assert first_event.neighbor == NodeID(1)
        assert known_neighbors0 == {NodeID(1)}

        second_event = queue0.get(block=True, timeout=0.5)
        assert isinstance(second_event, Recv)
        assert second_event.neighbor == NodeID(1)
        assert second_event.payload == b'Nice to meet you, Node 0 :)'

    finally:
        dut0.stop()
        dut1.stop()


SENDALL_PORT_0 = base_port.next()
SENDALL_PORT_1 = base_port.next()
SENDALL_PORT_2 = base_port.next()
SENDALL_PORT_3 = base_port.next()
SENDALL_PORT_4 = base_port.next()


SENDALL_NODES_AND_ENDPOINTS = [
    (NodeID(0), IPEndpoint(LOOPBACK, SENDALL_PORT_0)),
    (NodeID(1), IPEndpoint(LOOPBACK, SENDALL_PORT_1)),
    (NodeID(2), IPEndpoint(LOOPBACK, SENDALL_PORT_2)),
    (NodeID(3), IPEndpoint(LOOPBACK, SENDALL_PORT_3)),
    (NodeID(4), IPEndpoint(LOOPBACK, SENDALL_PORT_4)),
]


def test_sendall() -> None:
    '''One node sends to a bunch of neighbors.'''

    known_neighbors: Dict[NodeID, Set[NodeID]] = {}
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    (0, (LOOPBACK, SENDALL_PORT_0)),
                    (1, (LOOPBACK, SENDALL_PORT_1)),
                    (2, (LOOPBACK, SENDALL_PORT_2)),
                    (3, (LOOPBACK, SENDALL_PORT_3)),
                    (4, (LOOPBACK, SENDALL_PORT_4)),
                ]
            }
        }
    })

    dut: Dict[NodeID, SocketsCommunicator] = {}
    queue: Dict[NodeID, Queue] = {}
    for node_id in range(0, 5):
        node = NodeID(node_id)
        known_neighbors[node] = set()
        queue[node] = Queue()
        dut[node] = SocketsCommunicator(
            my_id=node,
            known_neighbors=known_neighbors[node],
            distributed=distributed)
        dut[node].start(queue=queue[node], timeout=0.5)

    for node_id in range(1, 5):
        node = NodeID(node_id)
        # We need to send an initial message to establish our neighborliness.
        dut[node].send([NodeID(0)], f'Hello, I am {node}.'.encode())

    try:
        # Read a NewNeighbor and a Recv for each neighbor.
        # The pairs of requests are randomly interleaved due to threading,
        # so we don't bother checking what they are.
        for _ in range(1, 5):
            _ = queue[NodeID(0)].get()
            _ = queue[NodeID(0)].get()

        # This is what we are testing.
        dut[NodeID(0)].send_all(payload=b'Hello, All Nodes!')

        # Confirm that the test worked.
        for node_id in range(1, 5):
            node = NodeID(node_id)

            first_event = queue[node].get(block=True, timeout=0.5)
            assert isinstance(first_event, NewNeighbor)
            assert first_event.neighbor == NodeID(0)
            assert known_neighbors[node] == {NodeID(0)}

            second_event = queue[node].get(block=True, timeout=0.5)
            assert isinstance(second_event, Recv)
            assert second_event.neighbor == NodeID(0)
            assert known_neighbors[node] == {NodeID(0)}
            assert second_event.payload == b'Hello, All Nodes!'

    finally:
        for node_id in range(0, 5):
            node = NodeID(node_id)
            dut[node].stop()


DYNAMIC_DISCOVERY_PORT_0 = base_port.next()
DYNAMIC_DISCOVERY_PORT_1 = base_port.next()
DYNAMIC_DISCOVERY_PORT_2 = base_port.next()
DYNAMIC_DISCOVERY_PORT_3 = base_port.next()
DYNAMIC_DISCOVERY_PORT_4 = base_port.next()


DYNAMIC_DISCOVERY_NODES_AND_ENDPOINTS = [
    (NodeID(0), IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_0)),
]

DYNAMIC_DISCOVERY_ENDPOINTS = [
    IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_0),
    IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_1),
    IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_2),
    IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_3),
    IPEndpoint(LOOPBACK, DYNAMIC_DISCOVERY_PORT_4),
]


def test_dynamic_discovery() -> None:
    '''One node sends to a bunch of neighbors.'''

    known_neighbors: Dict[NodeID, Set[NodeID]] = {}
    distributed = DistributedConfig(clause={
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    (0, (LOOPBACK, DYNAMIC_DISCOVERY_PORT_0)),
                    (1, (LOOPBACK, DYNAMIC_DISCOVERY_PORT_1)),
                    (2, (LOOPBACK, DYNAMIC_DISCOVERY_PORT_2)),
                    (3, (LOOPBACK, DYNAMIC_DISCOVERY_PORT_3)),
                    (4, (LOOPBACK, DYNAMIC_DISCOVERY_PORT_4)),
                ]
            }
        }
    })

    dut: Dict[NodeID, SocketsCommunicator] = {}
    queue: Dict[NodeID, Queue] = {}
    for node_id in range(0, 5):
        node = NodeID(node_id)
        known_neighbors[node] = set()
        queue[node] = Queue()
        dut[node] = SocketsCommunicator(
            my_id=node,
            known_neighbors=known_neighbors[node],
            distributed=distributed)
        dut[node].start(queue=queue[node], timeout=0.5)

    # This is the test. Everybody introduces themselves.
    for node_id in range(1, 5):
        node = NodeID(node_id)
        dut[node].send([NodeID(0)], f'Hello, I am {node}.'.encode())

    # Confirm that the test worked.
    try:
        # Read a NewNeighbor and a Recv for each neighbor.
        # The pairs of requests are randomly interleaved due to threading,
        # so we don't bother checking what they are.
        # This is needed to make sure every message has been recieved before
        #   checking known_neighbors.
        for _ in range(1, 5):
            _ = queue[NodeID(0)].get()
            _ = queue[NodeID(0)].get()

        assert known_neighbors[NodeID(0)] == {
            NodeID(1),
            NodeID(2),
            NodeID(3),
            NodeID(4),
        }

        known_endpoints = {
            dut[NodeID(0)].lookup_endpoint(node_id).port
            for node_id in known_neighbors[NodeID(0)]
        }
        assert known_endpoints == {
            DYNAMIC_DISCOVERY_PORT_1,
            DYNAMIC_DISCOVERY_PORT_2,
            DYNAMIC_DISCOVERY_PORT_3,
            DYNAMIC_DISCOVERY_PORT_4,
        }

    finally:
        for node_id in range(0, 5):
            node = NodeID(node_id)
            dut[node].stop()
